/*+ Util.cpp
 *
 ******************************************************************************
 *
 *                        Trimble Navigation Limited
 *                           645 North Mary Avenue
 *                              P.O. Box 3642
 *                         Sunnyvale, CA 94088-3642
 *
 ******************************************************************************
 *
 *    Copyright  2005 Trimble Navigation Ltd.
 *    All Rights Reserved
 *
 ******************************************************************************
 *
 * Description:
 *    This file implements the CUtil class.
 *            
 * Revision History:
 *    09-22-2005    Mike Priven
 *                  Cleaned up and adapted for outside distribution.
 *
 *    xx-xx-2001    Sean Zhi
 *                  Original
 *
 * Notes:
 *
-*/

/*---------------------------------------------------------------------------*\
 |                         I N C L U D E   F I L E S
\*---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "Const.h"
#include "Util.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/*---------------------------------------------------------------------------*\
 |                    M E T H O D   D E F I N I T I O N S
\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
Function:       LoadSrecFile

Description:    Parses an ASCII S-record file, extracts binary data, and loads
                into the internal buffer.
                
                **** IMPORTANT NOTES *******
                The data must be loaded into the memory buffer in a contigous
                manner. However, the S-records may not be sorted in the order
                of the lowest to highest loading address. Therefore, care
                must be taken when parsing the S-record file and loading data
                into the memory buffer. This function takes care of this issue
                by loading data at the offset into the buffer calculated as
                the difference between the loading address and the base 
                address, which is the lowest loading address.
                
                This function also assumes individual S-records are continuous 
                without any breaks.
                
Parameters:     fp  - a pointer to the ASCII S-record file
                buf - a pointer to memory where the binary data will be loaded
                      to.
                base - starting (base) address against which the records in the
                       file will be checked. This serves as a sanity check to 
                       make sure the specified file corresponds to the proper
                       settings (FW ROM file or Boot ROM file).

Return Value:   Size of the binary loadable data, in bytes
-----------------------------------------------------------------------------*/
long CUtil::LoadSrecFile (FILE* fp, unsigned char* buf, unsigned long base)
{
    char	      line[LINE_SIZE];
    long	      totalDataSize, offset;
    short	      dataSize, i;
    unsigned char *ptr;
    unsigned long temp;
    unsigned long address;
    
    // Get the size of the binary data and allocate heap space for it.
    if ((totalDataSize = GetSrecDataSize (fp)) < 0)
    {
        // S-rec file format problem
        return ERR_SREC_FORMAT;
    }
    
    // Default data byte is 0xFF to indicate that it should not be loaded
    // since it contains no usable data.
    memset (buf, 0xFF, totalDataSize);
    
    ptr = buf;
    
    // Rewind the file pointer to the beginning
    rewind (fp);
    
    // For each line in the ASCII S-record file, perform the following:
    while (fgets (line, LINE_SIZE, fp) != NULL) 
    {
        // Only S3 records are parsed. The type of record field is 'S3' (0x5333). 
        // The address field is intrepreted as a 4-byte address. The data field 
        // is composed of memory loadable data.
        if (line[1] != '3') continue;
        
        // Check the S-record data against the checksum. The function returns
        // the size of the loadable data bytes.
        if ((dataSize = CheckSrecCsum (line)) == -1)
        {
            // S-rec checksum incorrect
            return ERR_SREC_CHKSM; 
        }
        
        // Extract the loading address from the S-record
        if ((address = GetSrecAddr (line)) == INVALID_ADDR) 
        {
            free (ptr);
            ptr = NULL;
            // invalid S-rec address value
            return ERR_SREC_ADDR; 
        }
        
        // A sanity check to make sure the ROM type selected in Settings matches
        // the file being passed in.
        if (base > address)
        {
            free (ptr);
            ptr = NULL;
            // ROM type mismatch; base address doesn't match
            return ERR_SREC_BASE; 
        }
        
        // Get the offset into the memory buffer where the data will be loaded.
        // The offset is just the difference between the loading address of the
        // current S-record and the base loading address.
        offset = address - base;
        
        // A sanity check to make sure the ROM type selected in Settings matches
        // the file being passed in. The offset must be less than the total number
        // of data bytes, and it must not be negative.
        if ((offset > totalDataSize) || (offset < 0))
        {
            free (ptr);
            ptr = NULL;
            // S-rec file size mismatch
            return ERR_SREC_SIZE; 
        }
        
        // For each two ASCII characters in the S-record string representing a 
        // single data byte, extract and store into the memory buffer at the 
        // appropriate offset. Extracting starts in the string at the index 12 
        // (zero-based array access) and advances by 2 characters for a total 
        // number of individual data bytes in the string.
        for (i=0; i<dataSize; i++) 
        {
            sscanf (&line[12+i*2], "%02x", &temp);
            ptr[offset+i] = (unsigned char)temp;
        }
    }
    
    return totalDataSize;
}

/*-----------------------------------------------------------------------------
Function:       LoadSrecStr

Description:    Parses an ASCII S-record char string, extracts binary data, and loads
                into the internal buffer.
                
                **** IMPORTANT NOTES *******
                The data must be loaded into the memory buffer in a contigous
                manner. However, the S-records may not be sorted in the order
                of the lowest to highest loading address. Therefore, care
                must be taken when parsing the S-record file and loading data
                into the memory buffer. This function takes care of this issue
                by loading data at the offset into the buffer calculated as
                the difference between the loading address and the base 
                address, which is the lowest loading address.
                
                This function also assumes individual S-records are continuous 
                without any breaks.
                
Parameters:     fp  - a pointer to the ASCII S-record file
                buf - a pointer to memory where the binary data will be loaded
                      to.
                base - starting (base) address against which the records in the
                       file will be checked. This serves as a sanity check to 
                       make sure the specified file corresponds to the proper
                       settings (FW ROM file or Boot ROM file).

Return Value:   Size of the binary loadable data, in bytes
-----------------------------------------------------------------------------*/
long CUtil::LoadSrecStr (char* demon, unsigned char* buf, unsigned long base)
{
    char	      line[LINE_SIZE+10];
    short         demonIdx, sRecordIdx;
    long	      totalDataSize, offset;
    short	      dataSize, i;
    unsigned char *ptr;
    unsigned long temp;
    unsigned long address;
    
    if (demon == NULL)
    {
        return ERR_SREC_FORMAT;
    }
    
    // Get the size of the binary data and allocate heap space for it.
    if ((totalDataSize = GetSrecStrSize(demon)) < 0)
    {
        // S-rec file format problem
        return ERR_SREC_FORMAT;
    }
    
    // Default data byte is 0xFF to indicate that it should not be loaded
    // since it contains no usable data.
    memset (buf, 0xFF, totalDataSize);
    
    ptr = buf;
    
    // Rewind the file pointer to the beginning
    demonIdx = 0;
    
    // For each line in the ASCII S-record file, perform the following:
    while (demonIdx < totalDataSize)
    {
        // Assume the first byte is 'S'.
        line[0] = demon[demonIdx++];
        sRecordIdx = 1;
        while ((sRecordIdx < LINE_SIZE) && (demon[demonIdx] != 'S'))
        {
            line[sRecordIdx++] = demon[demonIdx++];
        }
        
        line[sRecordIdx] = 0;
        
        // Only S3 records are parsed. The type of record field is 'S3' (0x5333). 
        // The address field is intrepreted as a 4-byte address. The data field 
        // is composed of memory loadable data.
        if (line[1] != '3') continue;
        
        // Check the S-record data against the checksum. The function returns
        // the size of the loadable data bytes.
        if ((dataSize = CheckSrecCsum (line)) == -1)
        {
            // S-rec checksum incorrect
            return ERR_SREC_CHKSM; 
        }
        
        // Extract the loading address from the S-record
        if ((address = GetSrecAddr (line)) == INVALID_ADDR) 
        {
            free (ptr);
            ptr = NULL;
            // invalid S-rec address value
            return ERR_SREC_ADDR; 
        }
        
        // A sanity check to make sure the ROM type selected in Settings matches
        // the file being passed in.
        if (base > address)
        {
            free (ptr);
            ptr = NULL;
            // ROM type mismatch; base address doesn't match
            return ERR_SREC_BASE; 
        }
        
        // Get the offset into the memory buffer where the data will be loaded.
        // The offset is just the difference between the loading address of the
        // current S-record and the base loading address.
        offset = address - base;
        
        // A sanity check to make sure the ROM type selected in Settings matches
        // the file being passed in. The offset must be less than the total number
        // of data bytes, and it must not be negative.
        if ((offset > totalDataSize) || (offset < 0))
        {
            free (ptr);
            ptr = NULL;
            // S-rec file size mismatch
            return ERR_SREC_SIZE; 
        }
        
        // For each two ASCII characters in the S-record string representing a 
        // single data byte, extract and store into the memory buffer at the 
        // appropriate offset. Extracting starts in the string at the index 12 
        // (zero-based array access) and advances by 2 characters for a total 
        // number of individual data bytes in the string.
        for (i=0; i<dataSize; i++) 
        {
            sscanf (&line[12+i*2], "%02x", &temp);
            ptr[offset+i] = (unsigned char)temp;
        }
    }
    
    return totalDataSize;
}

/*-----------------------------------------------------------------------------
Function:       GetSrecStrSize

Description:    Returns the size of the loadable data in bytes.
                
Parameters:     fp  - a pointer to the ASCII S-record file

Return Value:   Size of the binary loadable data, in bytes
-----------------------------------------------------------------------------*/
long CUtil::GetSrecStrSize (char *demon)
{
    // return the length of the string
    return strlen(demon);
}

/*-----------------------------------------------------------------------------
Function:       GetSrecDataSize

Description:    Parses an ASCII S-record file and returns the size of the
                loadable data in bytes.
                
Parameters:     fp  - a pointer to the ASCII S-record file

Return Value:   Size of the binary loadable data, in bytes
-----------------------------------------------------------------------------*/
long CUtil::GetSrecDataSize (FILE* fp)
{
    char line[LINE_SIZE];
    long length = 0;
    long count;
    
    // Rewind the file pointer to the beginning
    rewind (fp);
    
    while (fgets (line, LINE_SIZE, fp) != NULL) 
    {
        // Only S3 records are parsed. The type of record field is 'S3' (0x5333). 
        // The address field is intrepreted as a 4-byte address. The data field 
        // is composed of memory loadable data.
        if (line[1] != '3') continue;
        
        // The 3rd and 4th characters in an S-record represent a byte 
        // containing a count of 2-character pairs in the rest of the record.
        if (sscanf (&line[2], "%02x", &count) == 0)
        {
            return -1;
        }
        else
        {
            // The number of loadable data bytes is the value of "count" in
            // the S-record minus 4 bytes for the address field and 1 byte
            // for the checksum field.
            length += (count - 5);
        }
    }
    
    return length;
}

/*-----------------------------------------------------------------------------
Function:       CheckSrecCsum

Description:    Parses a single ASCII S-record of type S3, computes the 
                checksum of data, compares to the stored checksum, and returns 
                the number of loadable data bytes or -1 if the checksum 
                verification failed.
                
Parameters:     str - a line containing a single S3 record

Return Value:   Number of the loadable data contained in the record, in bytes
                or -1 if checksum verification failed or an invalid S-record 
                format is detected
-----------------------------------------------------------------------------*/
short CUtil::CheckSrecCsum (char* str)
{
    unsigned long temp, count, idx, xsumSrec;
    unsigned char xsum;
    
    // The 3rd and 4th characters in an S-record represent a byte 
    // containing a count of 2-character pairs in the rest of the record.
    if (sscanf (&str[2], "%02x", &count) != 1)
    {
        return -1;
    }
    
    // Compute the checksum: ones complement of the sum of all bytes starting 
    // at count.
    for (idx = 2, xsum = 0; idx <= count*2; idx += 2) 
    {
        sscanf (&str[idx], "%02x", &temp);
        xsum += (unsigned char)temp;
    }
    xsum = ~xsum;
    
    // Extract the stored checksum from the S-record.
    if (sscanf (&str[2+count*2], "%02x", &xsumSrec) != 1)
    {
        return -1;
    }
    
    // Compare the stored checksum to the computed checksum.
    if (xsum == (unsigned char)xsumSrec)
    {
        // The return value is the number of loadable data bytes, so we need 
        // to subtract from the total byte count 4 bytes used for the address 
        // field and 1 byte used for the checksum.
        return (short)(count-5);
    }
    else
    {
        return -1;
    }
}

/*-----------------------------------------------------------------------------
Function:       GetSrecAddr

Description:    Parses a single ASCII S-record of type S3 and extracts the 
                loading address.
                
Parameters:     str - a line containing a single S3 record

Return Value:   Loading address or INVALID_ADDR if S-record format is invalid
-----------------------------------------------------------------------------*/
unsigned long CUtil::GetSrecAddr (char* str)
{
    unsigned long address;
    
    // Bytes 4-7 of the S3 type of S-record contain a 4-byte loading address.
    if (sscanf (&str[4], "%08x", &address) == 0)
    {
        return INVALID_ADDR;
    }
    else
    {
        return address;
    }
}

/*-----------------------------------------------------------------------------
Function:       GetForce2MonPkt

Description:    Loads the specified string with a TSIP packet "Force to Boot
                Monitor".

                NOTE: this packet is formatted using the TSIP protocol. It
                      is sent to the target operating in the normal TSIP output
                      mode to force it into the boot monitor mode which has
                      a different protocol format.
                
Parameters:     str - a pointer to a string to load the packet data into.

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil:: GetForce2MonPkt (char* str)
{
    short idx = 0;
    
    str[idx++] = 0x10;
    str[idx++] = 0x1E;
    str[idx++] = 0x4D;
    str[idx++] = 0x10;
    str[idx++] = 0x03;
    
    return idx;
}

/*-----------------------------------------------------------------------------
Function:       GetEnqPkt

Description:    Loads the specified string with a boot monitor packet ENQ.
                
Parameters:     str - a pointer to a string to load the packet data into.

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil::GetEnqPkt (char* str)
{
    short idx = 0;
    
    str[idx++] = ENQ;
    str[idx++] = (ENQ<<1);
    str[idx++] = (ENQ<<2);
    str[idx++] = (ENQ<<3);
    
    return idx;
}

/*-----------------------------------------------------------------------------
Function:       GetFwErasePkt

Description:    Loads the specified string with a boot monitor packet "Erase
                Firmware Section".
                
Parameters:     str - a pointer to a string to load the packet data into.

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil::GetFwErasePkt (char* str)
{
    short idx = 0;
    char  xsum = 0;
    
    str[idx++] = STX;
    str[idx++] = 0x00;
    str[idx++] = xsum = FW_ERASE_PKT_ID;
    str[idx++] = 0;
    str[idx++] = xsum;
    str[idx++] = ETX;
    
    return idx;
}

/*-----------------------------------------------------------------------------
Function:       GetFwProgramPkt

Description:    Loads the specified string with a boot monitor packet "Start
                Firmware Programming".
                
Parameters:     str - a pointer to a string to load the packet data into.
                size - total length of the Firmware data (in bytes)
                base - starting adress of the Firmware section in FLASH

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil::GetFwProgramPkt (char* str, char* data, short size, unsigned long base)
{
    short idx = 0, len;
    short xsum = 0;
    
    str[idx++] = STX;
    str[idx++] = 0x00;
    
    str[idx] = FLASH_PROGRAM_PKT_ID;
    xsum += str[idx++];
    
    str[idx] = (char)size+4;
    xsum += str[idx++];
    
    str[idx++] = GET_BYTE_3(base);
    xsum += GET_BYTE_3(base);
    str[idx++] = GET_BYTE_2(base);
    xsum += GET_BYTE_2(base);
    str[idx++] = GET_BYTE_1(base);
    xsum += GET_BYTE_1(base);
    str[idx++] = GET_BYTE_0(base);
    xsum += GET_BYTE_0(base);
    
    for (len = 0; len < size ; len++) 
    {
        xsum += data[len];
        str[idx++] = data[len];
    }
    
    str[idx++] = (char)xsum;
    str[idx++] = ETX;
    
    return idx;
}

/*-----------------------------------------------------------------------------
Function:       GetDemonProgramPkt

Description:    Loads the specified string with a boot monitor packet "Start
                Demon Programming".
                
Parameters:     str - a pointer to a string to load the packet data into.
                size - total length of the Demon data (in bytes)
                base - starting adress of the Demon section in FLASH

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil::GetDemonProgramPkt (char* str, char* data, short size, unsigned long base)
{
    short idx = 0;
    short len;
    short xsum = 0;
    
    str[idx++] = STX;
    str[idx++] = 0x00;
    
    str[idx] = DOWNLOAD_DATA_PKT_ID;
    xsum += str[idx++];
    
    str[idx] = (char)size+4;
    xsum += str[idx++];
    
    str[idx++] = GET_BYTE_3(base);
    xsum += GET_BYTE_3(base);
    str[idx++] = GET_BYTE_2(base);
    xsum += GET_BYTE_2(base);
    str[idx++] = GET_BYTE_1(base);
    xsum += GET_BYTE_1(base);
    str[idx++] = GET_BYTE_0(base);
    xsum += GET_BYTE_0(base);
    
    for (len = 0; len < size ; len++) 
    {
        xsum += data[len];
        str[idx++] = data[len];
    }
    
    str[idx++] = (char)xsum;
    str[idx++] = ETX;
    
    return idx;
}

/*-----------------------------------------------------------------------------
Function:       GetStartDemonPkt

Description:    Loads the specified string with a boot monitor packet "Start 
                Demon".
                
Parameters:     str - a pointer to a string to load the packet data into.

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil:: GetStartDemonPkt (char* str)
{
    short idx = 0;
    char  xsum = 0;
    
    str[idx++] = STX;
    str[idx++] = 0x00;
    str[idx++] = xsum = RUNCODEAT_PKT_ID;
    str[idx++] = 0x04;  // size
    xsum += 0x04;
    str[idx++] = GET_BYTE_3(DEMON_START_ADDR);
    xsum += GET_BYTE_3(DEMON_START_ADDR);
    str[idx++] = GET_BYTE_2(DEMON_START_ADDR);
    xsum += GET_BYTE_2(DEMON_START_ADDR);
    str[idx++] = GET_BYTE_1(DEMON_START_ADDR);
    xsum += GET_BYTE_1(DEMON_START_ADDR);
    str[idx++] = GET_BYTE_0(DEMON_START_ADDR);
    xsum += GET_BYTE_0(DEMON_START_ADDR);
    str[idx++] = xsum;
    str[idx++] = ETX;
    
    return idx;
}


/*-----------------------------------------------------------------------------
Function:       GetDemonSetPortPkt

Description:    Loads the specified string with a demon packet "Set Port Baud".
                
Parameters:     str - a pointer to a string to load the packet data into.
                baud - baud rate to set.

Return Value:   Total size of the packet, in bytes
-----------------------------------------------------------------------------*/
short CUtil:: GetDemonSetPortPkt (char* str, unsigned long baud)
{
    short idx = 0;
    char  xsum = 0, rate;
    
    switch (baud)
    {
        case 115200:
            rate = 0x0E;
            break;
        case 57600:
            rate = 0x0D;
            break;
        case 38400:
            rate = 0x0C;
            break;
        case 9600:
        default:
            rate = 0x0B;
            break;
    }
    
    str[idx++] = STX;
    str[idx++] = 0x00;
    str[idx++] = xsum = BAUD_CONTROL_PKT_ID;
    str[idx++] = 1;
    xsum += 1;
    str[idx++] = rate;
    xsum += rate;
    str[idx++] = xsum;
    str[idx++] = ETX;
    
    return idx;
}

